/*----------------------------------------------------------------------------
    ChucK Concurrent, On-the-fly Audio Programming Language
      Compiler and Virtual Machine

    Copyright (c) 2004 Ge Wang and Perry R. Cook.  All rights reserved.
      http://chuck.cs.princeton.edu/
      http://soundlab.cs.princeton.edu/

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
    U.S.A.
-----------------------------------------------------------------------------*/

//-----------------------------------------------------------------------------
// name: chuck_ugen.h
// desc: chuck unit generator interface
//
// authors: Ge Wang (gewang@cs.princeton.edu)
//          Perry R. Cook (prc@cs.princeton.edu)
// date: Spring 2004
//-----------------------------------------------------------------------------
#ifndef __CHUCK_UGEN_H__
#define __CHUCK_UGEN_H__

#include "chuck_def.h"
#include "chuck_oo.h"
#include <string>
#include <vector>
#include <map>
using namespace std;


// forward reference
struct Chuck_UGen;
struct Chuck_UGen_Info;
struct Chuck_Info_Param;


// dynamic linking unit generator interface prototypes
extern "C" {
typedef void *   (* f_ctor)( t_CKTIME now );
typedef void     (* f_dtor)( t_CKTIME now, void * data );
typedef t_CKBOOL (* f_tick)( t_CKTIME now, void * data, SAMPLE in, SAMPLE * out );
typedef void     (* f_ctrl)( t_CKTIME now, void * data, void * value );
typedef t_CKBOOL (* f_pmsg)( t_CKTIME now, void * data, const char * msg, void * value );
}




//-----------------------------------------------------------------------------
// name: struct Chuck_Info_Param
// dsec: ugen info param
//-----------------------------------------------------------------------------
struct Chuck_Info_Param
{
    string type;  // see above
    string name;  // name of param
    f_ctrl addr;  // addr

    // constructor
    Chuck_Info_Param( const string & t = "", const string & n = "", f_ctrl a = NULL )
    { type = t; name = n; addr = a; }
    
    // copy constructor
    Chuck_Info_Param( const Chuck_Info_Param & p )
        : type(p.type), name(p.name), addr(p.addr) { }
};




//-----------------------------------------------------------------------------
// name: struct Chuck_UGen_Info
// dsec: ugen info
//-----------------------------------------------------------------------------
struct Chuck_UGen_Info
{
    // info
    string name;
    vector<Chuck_Info_Param> param_list;
    map<string, Chuck_Info_Param> param_map;
    f_ctor ctor;
    f_dtor dtor;
    f_tick tick;
    f_pmsg pmsg;
    
    // optional
    string parent;
    t_CKUINT min_src;
    t_CKUINT max_src;
    int linepos;
    
    // set
    void set( f_ctor c, f_dtor d, f_tick t, f_pmsg p )
    { ctor = c; dtor = d; tick = t; pmsg = p; }
    
    // add
    void add( f_ctrl c, const string & t, const string & n )
    { param_list.push_back( Chuck_Info_Param( t, n, c ) );
      param_map[n] = param_list[param_list.size()-1]; }

    // contructor
    Chuck_UGen_Info( const string & n, t_CKUINT min = 0, t_CKUINT max = 0xffffffff )
    {
        name = n;
        parent = "";
        min_src = min;
        max_src = max;
        set( NULL, NULL, NULL, NULL );
        linepos = 0;
    }
};


// op mode
#define UGEN_OP_PASS    -1
#define UGEN_OP_STOP    0
#define UGEN_OP_TICK    1


//-----------------------------------------------------------------------------
// name: class Chuck_UGen
// dsec: ugen base
//-----------------------------------------------------------------------------
struct Chuck_UGen : public Chuck_Object
{
public:
    Chuck_UGen( );
    void init();
    
public: // src
    t_CKBOOL add( Chuck_UGen * src );
    t_CKBOOL remove( Chuck_UGen * src );
    t_CKBOOL set_max_src( t_CKUINT num );
    t_CKUINT get_num_src( );
    t_CKUINT system_tick( t_CKTIME now );

public: // interface
    f_ctor ctor;
    f_dtor dtor;
    f_tick tick;
    f_pmsg pmsg;

public: // data
    vector<Chuck_UGen *> m_src_list;
    t_CKUINT m_num_src;
    t_CKUINT m_max_src;
    t_CKTIME m_time;
    t_CKBOOL m_valid;
    SAMPLE m_sum;
    SAMPLE m_current;
    SAMPLE m_gain;
    int m_op;

public: // dynamic linking client data
    void * state;
};




#endif
